# Multi Qubit Correction (MQC) for Quantum Annealers  
by Ramin Ayanzadeh, John Dorband, Milton Halem and Tim Finin   
**Publisher:** Nature Publishing Group | 
**Journal:** Scientific Reports (In Press) | 
**Year:** 2021  
[AR14: https://arxiv.org/abs/2010.00115](https://arxiv.org/abs/2010.00115)  

<pre></pre>


## About MQC 
MQC is a classical postprocessing method for improving the quality of drawn samples by quantum annealers (like the D-Wave quantum processors), and any Ising processing unit (IPU). 
MQC receives a sample-set and coefficients of an Ising model as input, 
and (in most cases) result in a new/synthetic sample that is notably better than all input samples. 
MQC has outperformne recent software/hardware advancements such as applying 
spin-reversal transforms, increased inter-sample delay between successive reads/samples, applying steepest descent method (as a postprocessing approach) to drawn samples, reverse annealing, etc. 
It is worth noting that MQC requires significantly fewer samples.      

<pre></pre>

# How to use MQC 

The ***.zip*** file includes *testing_MQC* notebook/Python files that explains how one can use MQC. 

## Prerequisits 

To use MQC, following packages are needed:
- [networkx](https://pypi.org/project/networkx/)

- [dwave-ocean-sdk](https://pypi.org/project/dwave-ocean-sdk/) (optional/recommended)--This package is required to running the provided *testing_MQC* notebook and Python files.

<pre></pre>

## Example: 
\# Given linear and quadratic coefficients of an Ising model, denoted by h and J, respectively  

**from** MQC **import**  \*  
**import** neal   
**import** dimod   
sampler = neal.SimulatedAnnealingSampler()  

\#Drawing 10,000 samples on a classical computer can take several minutes  
SA_samples= sampler.sample_ising(h, J, num_reads=10000)  
e_SA=SA_samples.first.energy  

\#Applying MQC to first 100 samples, attained by SA sampler   
SA_sliced_samples=SA_samples.slice(0, 100, sorted_by=None)  
print(f'Size of the sliced sampleset: {len(SA_sliced_samples.samples())}')  

s_SA_MQC=MQC(SA_sliced_samples.samples(sorted_by=None), h, J)  
e_SA_MQC=dimod.utilities.ising_energy(s_SA_MQC, h, J)

print(f'Lowest Energy value of 10,000 samples attained by Simulated Annealing sampler: {e_SA}')    
print(f'Energy value of the sample attained by applying MQC to only first 100 samples: {e_SA_MQC}')  


## Output  
For a random Ising problem on Chimera graph (i.e., hardware topology of the D-Wave 2000Q QPU), where all coefficients were drawn randomly followed by a standard normal distribution: 

    Size of the sliced sampleset: 100  
    Lowest Energy value of 10,000 samples attained by Simulated Annealing sampler: -3843.557576218682  
    Energy value of the sample attained by MQC: -3863.415378307025

<pre></pre>

For more (comprehensive) example, please visit the *testing_MQC* Jupyter notebook or Python file.  

<pre></pre>
**Note**: C/C++ implementation of MQC can provide up to 100x speed-up compared with this *Python* implementation.
