# This Python script file has been prepared to demonstrate that 
# applying Multi-Qubit Correction method (MQC) 
# by Ramin Ayanzadeh, John Dorband, Milton Halem and Tim Finin 
# https://arxiv.org/abs/2010.00115
# to (notably fewer samples)) attained by quantum annealers 
# can result in a sample with significantly lower energy value 


# # Loading an example Ising problem along with 1,000 samples attained by a D-Wave 2000Q Quantum Annealer

import pickle

pkl_filename='example.pkl'
with open(pkl_filename, 'rb') as pkl:
        obj=pickle.load(pkl)
print(obj.keys())

h=obj['h']
J=obj['J']
QA_samples=obj['samples']



# # Finding the Lowest Energy Value of Samples Attained by D-Wave QA

print(f'{len(QA_samples)} samples have been drawn using a D-Wave 2000Q quantum annealer')
e_QA=QA_samples.first.energy
print(f'Lowest energy value attained by QA: {e_QA}')


# # Applying Multi-Qubit Correction (MQC) to first 100 samples 

import MQC
print('Aplying Multi-Qubit Correction (MQC) method to first 100 samples attained by a D-Wave 2000Q quantum annealer')
sliced_samples=QA_samples.slice(0, 100, sorted_by=None)
print(f'Size of the sliced sampleset: {len(sliced_samples.samples())}')

s_MQC=MQC(sliced_samples.samples(sorted_by=None), h, J)

#finding the Ising energy value of the sample fount by MQC
import dimod 
e_MQC=dimod.utilities.ising_energy(s_MQC, h, J)
print(f'Energy value of the sample attained by MQC: {e_MQC}')


# # Comparing QA with MQC

print('Ising energy values:')
# QA has drawn 10,000 samples 
print(f'QA:  {e_QA}')

#MQC has been applied to only first 100 samples attained by a QA
print(f'MQC: {e_MQC}')


#########################################
#########################################
# # Another example: Applying MQC to samples attained by Simulated Annealing sampler 

import neal
sampler = neal.SimulatedAnnealingSampler()

#drawing 10,000 samples on a classical computer can take several minutes
print('Drawing 10,000 samples on a classical computer can take several minutes')

SA_samples= sampler.sample_ising(h, J, num_reads=10000)
e_SA=SA_samples.first.energy
print(f'Lowest Energy value of 10,000 samples attained by Simulated Annealing sampler: {e_SA}')


# # Applying Multi-Qubit Correction (MQC) to first 100 samples  attained by Simulated Annealing sampler

#Applying MQC to first 100 samples, attained by SA sampler 
print('Aplying Multi-Qubit Correction (MQC) method to first 100 samples attained by Simulated Annealing sampler')

SA_sliced_samples=SA_samples.slice(0, 100, sorted_by=None)
print(f'Size of the sliced sampleset: {len(SA_sliced_samples.samples())}')

s_SA_MQC=MQC(SA_sliced_samples.samples(sorted_by=None), h, J)

#finding the Ising energy value of the sample fount by MQC
import dimod 
e_SA_MQC=dimod.utilities.ising_energy(s_SA_MQC, h, J)
print(f'Energy value of the sample attained by MQC: {e_SA_MQC}')


# # Comparing QA, SA and MQC

# QA has drawn 10,000 samples 
# SA has drawn 10,000 samples 
# MQC has been applied to only first 100 samples 

print('QA has drawn 10,000 samples') 
print('SA has drawn 10,000 samples') 
print('MQC has been applied to only first 100 samples')
print('------------------------')
print('Ising energy values:')
print(f'QA\t:{e_QA}')
print(f'SA\t:{e_SA}')
print(f'QA+MQC\t:{e_MQC}')
print(f'SA+MQC\t:{e_SA_MQC}')